/**************************************************************

fitsfun.c (C-Munipack project)
FITS frames manipulating functions
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "comfun.h"
#include "fitsfun.h"

/* Returns nonzero if the file is of FITS format */
int fits_test(FILE *f)
{
  int size;
  char buf[16];

  fseek(f,0,SEEK_END);
  size = ftell(f);
  rewind(f);
  /* Try FITS format */
  if (size>2880) {
//  if ((size>0)&&(((size%FITSSIZE)==0)||(size==200256)) {
// AlphaMaxi software produces FITS of invalid size (=200256 bytes)! :-(
	  if (fread(buf,10,1,f)) {
	 	  if (strstr(buf,"SIMPLE  =")==buf)
	 	      return 1;
      }
  }
  return 0;
}

/* Reads image resolution, number of samples per pixel and number of bits per sample */
int fits_gpar(fitsfile *fits, int *width, int *height, int *bitcount)
{
	int status = 0, simple, extend, naxis, bitpix;
    long pcount, gcount, naxes[2];

    ffghpr(fits,2,&simple,&bitpix,&naxis,naxes,&pcount,&gcount,&extend,&status);
    if (status==0) {
		if (width) *width = naxes[0];
		if (height) *height = naxes[1];
		if (bitcount) *bitcount = bitpix;
	}
	return status;
}

/* Reads date */
static int getdatetime(fitsfile *fits, int *yr, int *mo, int *dy, 
		int *hr, int *mi, int *se, int *ms, int *end)
{
	int x, status, date_ok, time_ok;
	char datestr[FLEN_VALUE], commstr[FLEN_COMMENT], timestr[FLEN_VALUE];
	long ltime;
	double dtime;

	*yr = *mo = *dy = 0;
	*hr = *mi = *se = *ms = 0;
	*end = 0;

	status = 0;
	ffgkys(fits,"DATE-OBS",datestr,commstr,&status);
    if (status!=0) { strcpy(datestr,""); status=0; }
    if (strlen(datestr)==0) {
	    // D02E-OBS= 'dd/mm/yy'
	    ffgkys(fits,"D02E-OBS",datestr,NULL,&status);
        if (status!=0) { strcpy(datestr,""); status=0; }
    }
    if (strlen(datestr)==0) {
	    // DATE-BEG= '2005-08-24T20:19:03'
		ffgkys(fits,"DATE-BEG",datestr,NULL,&status);
    	if (status!=0) { strcpy(datestr,""); status=0; }
    }
    
    date_ok = 0;
    if ((strlen(datestr)==8)&&(datestr[2]=='/')&&(datestr[5]=='/')) {
	    // dd/mm/yy
		*yr = 2000 + atoi(datestr+6);
		*mo = atoi(datestr+3);
		*dy = atoi(datestr);
		date_ok = 1;
    } else
    if (strlen(datestr)==10 && datestr[2]=='/' && datestr[5]=='/') {
    	// dd/mm/yyyy
		*yr = atoi(datestr+6);
		*mo = atoi(datestr+3);
		*dy = atoi(datestr);
		date_ok = 1;
    } else 
	if (strlen(datestr)>=10 && datestr[4]=='-' && datestr[7]=='-') {
		// yyyy-mm-dd
        *yr = atoi(datestr);
        *mo = atoi(datestr+5);
        *dy = atoi(datestr+8);
		date_ok = 1;
    } else 
	if (strlen(datestr)>=10 && datestr[4]=='/' && datestr[7]=='/') {
		// yyyy/mm/dd
        *yr = atoi(datestr);
        *mo = atoi(datestr+5);
        *dy = atoi(datestr+8);
		date_ok = 1;
	}
	
	status = 0;
	ffgkys(fits,"TIME-OBS",timestr,NULL,&status);
	if (status!=0) { strcpy(timestr,""); status=0; }
    if (strlen(timestr)==0) {
        ffgkys(fits,"UT",timestr,NULL,&status);
        if (status!=0) { strcpy(timestr,""); status=0; }
    }
    if (strlen(timestr)==0) {
	    // DATE-BEG= '2005-08-24T20:19:03'
		ffgkys(fits,"DATE-BEG",datestr,NULL,&status);
    	if (status!=0) { strcpy(datestr,""); status=0; }
		if ((strlen(datestr)>=18)&&(datestr[10]=='T')) {
  			datestr[10] = 0;
  			strcpy(timestr,datestr+11);
		}
    }
    if (strlen(timestr)==0) {
        // DATE-OBS= '2004-03-20T22:55:28.014' / GMT START OF EXPOSURE
		ffgkys(fits,"DATE-OBS",datestr,NULL,&status);
		if (status!=0) { strcpy(datestr,""); status=0; }
    	if ((strlen(datestr)>=18)&&(datestr[10]=='T')) {
  			datestr[10] = 0;
  			strcpy(timestr,datestr+11);
		}
    }
    if (strlen(timestr)==0) {
	    // TM-START=  1.3079446944444E+01 / [Corrected] UT obs'n start, real hours         
	    ffgkyd(fits,"TM-START",&dtime,NULL,&status);
	    if (status==0) {
		    ltime = (long)(dtime*3600000.0+0.5);
		    sprintf(timestr, "%02ld:%02ld:%02ld.%03ld", (ltime / 3600000L), (ltime / 60000L) % 60, 
		    	(ltime / 1000L) % 60, (ltime % 1000L));
	    } else {
		    strcpy(timestr,""); status=0; 
		}
    }
    if (strlen(timestr)==0) {
		// REC-STRT= '13:04:46'           / Recorded UT time of observation start
        ffgkys(fits,"REC-STRT",timestr,NULL,&status);
        if (status!=0) { strcpy(timestr,""); status=0; }
	}
    if (strlen(timestr)==0) {
	    // UT_END  = '19:33:59'				/ PMIS Image Processing Software
		ffgkys(fits,"UT_END",timestr,NULL,&status);
		if (status!=0) { strcpy(timestr,""); status=0; }
		else *end = 1;
    }
    if (strlen(timestr)==0) {
        ffgkys(fits,"UT-START",timestr,NULL,&status);
        if (status!=0) { strcpy(timestr,""); status=0; }
    }
    
    time_ok = 0;
	if ((strlen(timestr)>=8)&&(timestr[2]==':')&&(timestr[5]==':')) {
        *hr = atol(timestr);
        *mi = atol(timestr+3);
        x = (int)(1000.0*atof(timestr+6)+0.5);
        *se = x/1000;
        *ms = x%1000;
		time_ok = 1;
	}
	
	return (date_ok && time_ok ? 0 : MPK_INVALID_DATE);
}

/* Reads right ascension */
int fits_getra(fitsfile *fits, int *hr, int *mi, int *se)
{
	int status = 0;
	char rastr[FLEN_VALUE], commstr[FLEN_COMMENT];

	*hr = *mi = *se = 0;

	ffgkys(fits,"RA", rastr, commstr, &status);
    if (status!=0) { 
	    status = 0;
	    ffgkys(fits, "OBJCTRA", rastr, commstr, &status);
        if (status!=0) {
	        return MPK_INVALID_OBJRA;
        }
    }
       
    if ((strlen(rastr)>=8)&&(rastr[2]==':')&&(rastr[5]==':')) {
	    // hh:mm:ss.sss
		*hr = atoi(rastr);
		*mi = atoi(rastr+3);
		*se = (int) (atof(rastr+6)+0.5);
		return 0;
    }
    if ((strlen(rastr)>=7)&&(rastr[1]==':')&&(rastr[4]==':')) {
	    // h:mm:ss.sss
		*hr = atoi(rastr);
		*mi = atoi(rastr+2);
		*se = (int) (atof(rastr+5)+0.5);
		return 0;
    }
    if ((strlen(rastr)>=8)&&(rastr[2]==' ')&&(rastr[5]==' ')) {
	    // hh mm ss.sss
		*hr = atoi(rastr);
		*mi = atoi(rastr+3);
		*se = (int) (atof(rastr+6)+0.5);
		return 0;
    }
    if ((strlen(rastr)>=7)&&(rastr[1]==' ')&&(rastr[4]==' ')) {
	    // hh mm ss.sss
		*hr = atoi(rastr);
		*mi = atoi(rastr+2);
		*se = (int) (atof(rastr+5)+0.5);
		return 0;
    }
    
	return MPK_INVALID_OBJRA;
}

/* Reads declination */
int fits_getdec(fitsfile *fits, int *deg, int *mi, int *se)
{
	int status = 0;
	char decstr[FLEN_VALUE], commstr[FLEN_COMMENT];

	*deg = *mi = *se = 0;

	ffgkys(fits,"DEC", decstr, commstr, &status);
    if (status!=0) { 
	    status = 0;
	    ffgkys(fits, "OBJCTDEC", decstr, commstr, &status);
        if (status!=0) {
	        return MPK_INVALID_OBJDEC;
        }
    }
       
    if ((strlen(decstr)>=7)&&(decstr[1]==':')&&(decstr[4]==':')) {
	    // d:mm:ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+2);
		*se = (int) (atof(decstr+5)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=8)&&(decstr[1]=='+'||decstr[1]=='-')&&(decstr[2]==':')&&(decstr[5]==':')) {
	    // -d:mm:ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+3);
		*se = (int) (atof(decstr+6)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=8)&&(decstr[2]==':')&&(decstr[5]==':')) {
	    // dd:mm:ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+3);
		*se = (int) (atof(decstr+6)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=9)&&(decstr[1]=='+'||decstr[1]=='-')&&(decstr[3]==':')&&(decstr[6]==':')) {
	    // -dd:mm:ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+4);
		*se = (int) (atof(decstr+7)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=9)&&(decstr[3]==':')&&(decstr[6]==':')) {
	    // ddd:mm:ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+4);
		*se = (int) (atof(decstr+7)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=10)&&(decstr[1]=='+'||decstr[1]=='-')&&(decstr[4]==':')&&(decstr[7]==':')) {
	    // -ddd:mm:ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+5);
		*se = (int) (atof(decstr+8)+0.5);
		return 0;
    }
    
    if ((strlen(decstr)>=7)&&(decstr[1]==' ')&&(decstr[4]==' ')) {
	    // d mm ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+2);
		*se = (int) (atof(decstr+5)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=8)&&(decstr[1]=='+'||decstr[1]=='-')&&(decstr[2]==' ')&&(decstr[5]==' ')) {
	    // -d mm ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+3);
		*se = (int) (atof(decstr+6)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=8)&&(decstr[2]==' ')&&(decstr[5]==' ')) {
	    // dd mm ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+3);
		*se = (int) (atof(decstr+6)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=9)&&(decstr[1]=='+'||decstr[1]=='-')&&(decstr[3]==' ')&&(decstr[6]==' ')) {
	    // -dd mm ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+4);
		*se = (int) (atof(decstr+7)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=9)&&(decstr[3]==' ')&&(decstr[6]==' ')) {
	    // ddd mm ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+4);
		*se = (int) (atof(decstr+7)+0.5);
		return 0;
    }
    if ((strlen(decstr)>=10)&&(decstr[1]=='+'||decstr[1]=='-')&&(decstr[4]==' ')&&(decstr[7]==' ')) {
	    // -ddd mm ss.sss
		*deg = atoi(decstr);
		*mi = atoi(decstr+5);
		*se = (int) (atof(decstr+8)+0.5);
		return 0;
    }
    
	return MPK_INVALID_OBJDEC;
}


int fits_gfilter(fitsfile *fits, char *filter)
{
	int stat = 0;
	ffgkys(fits, "FILTER", filter, NULL, &stat);
	if (stat==0) return 0;
	stat = 0;
	ffgkys(fits, "FILTERS", filter, NULL, &stat);
	if (stat==0) return 0;
	strcpy(filter, "");
	return stat;
}

int fits_gobject(fitsfile *fits, char *object)
{
	int stat = 0;
	ffgkys(fits, "OBJECT",  object, NULL, &stat);
	if (stat!=0) strcpy(object, "");
	return stat;
}

int fits_gtemp(fitsfile *fits, double *temp)
{
	char buf[FLEN_VALUE];
	int stat = 0;
	ffgkyd(fits, "TEMPERAT", temp, NULL, &stat);
	if (stat==0) return 0;
	stat = 0;
	ffgkyd(fits, "CCD-TEMP", temp, NULL, &stat);
	if (stat==0) return 0;
	stat = 0;
	ffgkys(fits, "TEMP", buf, NULL, &stat);
	if (stat==0) { *temp = atof(buf); return 0; }
	*temp=999.99;
	return stat;
}

int fits_gexp(fitsfile *fits, double *exp)
{
	char buf[FLEN_VALUE];
	int stat = 0;
	ffgkyd(fits, "EXPTIME", exp, NULL, &stat);
	if (stat==0) return 0;
	stat = 0;
	ffgkys(fits, "EXPOSURE", buf, NULL, &stat);
	if (stat==0) { *exp = atof(buf); return 0; }
	*exp=0.0;
	return stat;
}

int fits_gblack(fitsfile *fits, double *blacklevel)
{
	char buf[FLEN_VALUE];
	int stat = 0;
	
	ffgkyd(fits, "BACKGRND", blacklevel, NULL, &stat);
	if (stat==0) return 0;
	stat = 0;
	ffgkyd(fits, "CBLACK", blacklevel, NULL, &stat);
	if (stat==0) return 0;
	stat = 0;
	ffgkys(fits, "DATAMIN", buf, NULL, &stat);
	if (stat==0) { *blacklevel = atof(buf); return 0; }
	*blacklevel=100.0;
	return stat;
}

int fits_grange(fitsfile *fits, double *range)
{
	char buf[FLEN_VALUE];
	int stat = 0;
	double blacklevel, whitelevel;
	
	ffgkyd(fits, "RANGE", range, NULL, &stat);
	if (stat==0) return 0;
	stat = 0;
	ffgkyd(fits, "CBLACK", &blacklevel, NULL, &stat);
	if (stat==0) { 
		ffgkyd(fits, "CWHITE", &whitelevel, NULL, &stat);
		if (stat==0) { 
			*range = (whitelevel-blacklevel); 
			return 0; 
		}
	}
	stat = 0;
	ffgkys(fits, "DATAMIN", buf, NULL, &stat);
	if (stat==0) { 
		blacklevel = atof(buf); 
		ffgkys(fits, "DATAMAX", buf, NULL, &stat);
		if (stat==0) { 
			whitelevel = atof(buf);
			*range = (whitelevel-blacklevel); 
			return 0;
		}
	}
	*range=100.0;
	return stat;
}

int fits_gfravg(fitsfile *fits, int *fravg)
/* Get number of frames averaged */
{
	int stat = 0; 
	long val;
	
	ffgkyj(fits, "FR_AVG", &val, NULL, &stat);
	if (stat==0) { 
		*fravg = val; 
		return 0; 
	}
	return stat;
}	

void fits_flipimage(fitsfile *fits, int flipx, int flipy, int *stat)
/* Flipping image */
{
  int nx, ny, simple, extend, bitpix, naxis, anyf, i, j, ii, jj;
  long pcount, gcount, naxes[2];
  double *ccd1, *ccd2;

  ffghpr(fits,2,&simple,&bitpix,&naxis,naxes,&pcount,&gcount,&extend,stat);
  nx  = naxes[0];
  ny  = naxes[1];
  ccd1 = (double*) malloc(nx*ny*sizeof(double));
  ccd2 = (double*) malloc(nx*ny*sizeof(double));
  ffg2dd(fits,1,0,nx,nx,ny,ccd1,&anyf,stat);
  for (i=0;i<nx;i++) {
      for (j=0; j<ny; j++) {
          if (flipx) ii = nx-i-1; else ii = i;
		  if (flipy) jj = ny-j-1; else jj = j;
	      ccd2[i+j*nx] = ccd1[ii+jj*nx];
      }
  }
  ffp2dd(fits,1,nx,nx,ny,ccd2,stat);
  free(ccd1);
  free(ccd2);
}

/* Updates date and time in FITS file */
int fits_setdatetime(fitsfile *fits, const char *datestr, const char *timestr)
{
  int stat = 0;
  if (datestr) ffukys(fits,"DATE-OBS", (char*) datestr, "UT DATE OF START", &stat);
  if (timestr) ffukys(fits,"TIME-OBS", (char*) timestr, "UT TIME OF START", &stat);
  return stat;
}

/* Updates date and time in FITS file */
int fits_setdt(fitsfile *fits, int yr, int mo, int dy, int hr, int mi, int se, int ms)
{
  char datestr[64], timestr[64];
  sprintf(datestr, "%04d-%02d-%02d", yr, mo, dy);
  sprintf(timestr, "%02d:%02d:%02d.%03d", hr, mi, se, ms);
  return fits_setdatetime(fits, datestr, timestr);
}

/* Updates date and time in FITS file */
int fits_setjd(fitsfile *fits, double jd)
{
	int yr, mo, dy, hr, mi, se, ms;
	cmpack_decodejd(jd, &yr, &mo, &dy, &hr, &mi, &se, &ms);
	return fits_setdt(fits, yr, mo, dy, hr, mi, se, ms);
}

int fits_getdatetime(fitsfile *fits, char *datestr, char *timestr)
{
	double jd;
	int yr, mo, dy, hr, mi, se, ms;
	
	if (fits_getjd(fits, &jd)==0) {
		cmpack_decodejd(jd, &yr, &mo, &dy, &hr, &mi, &se, &ms);
  		if (datestr) sprintf(datestr, "%04d-%02d-%02d", yr, mo, dy);
  		if (timestr) sprintf(timestr, "%02d:%02d:%02d.%03d", hr, mi, se, ms);
  		return 0;
	} else {
  		if (datestr) datestr[0] = '\0'; 
  		if (timestr) timestr[0] = '\0';
  		return MPK_INVALID_DATE;
	}
}

int fits_getjd(fitsfile *fits, double *jd)
{
	double tjd, exp;
	int yr, mo, dy, hr, mi, se, ms, end;
	
	if (getdatetime(fits, &yr, &mo, &dy, &hr, &mi, &se, &ms, &end)==0) {
		tjd = cmpack_encodejd(yr, mo, dy, hr, mi, se, ms);
		if (end) {
			if (fits_gexp(fits, &exp)==0) {
				if (jd) *jd = tjd - exp/86400.0;
				return 0;
			} else {
				if (jd) *jd = 0.0;
				return MPK_INVALID_DATE;
			}
		} else {
			if (jd) *jd = tjd;
			return 0;
		}
	} else {
		if (jd) *jd = 0;
		return MPK_INVALID_DATE;
	}
}
