/**************************************************************

imglist.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>

#include "imglist.h"

typedef struct ImageType
{
  void *image; 
  long *lextra;
  double *dextra;
  struct ImageType *next;	
} ImageType;

typedef struct ImageList
{
	ImageType *first, *last, *current; 
	int       count;
	int       extra_long, extra_dbl;
} ImageList;

tImgList ImgList_Create(int extra_long, int extra_double)
{
	ImageList *list = (ImageList*) malloc(sizeof(ImageList));
	if (list) {
		memset(list, 0, sizeof(ImageList));
		list->extra_long = extra_long;
		list->extra_dbl  = extra_double;
	}
	return (tImgList) list;
}

void ImgList_Clear(tImgList lst)
{
	ImageType *next, *ptr;
	ImageList *list	= (ImageList*) lst;
	
	if (list) {
	 	ptr = list->first;
 		while (ptr) {
     		next = ptr->next;
     		if (ptr->lextra) free(ptr->lextra);
     		if (ptr->dextra) free(ptr->dextra);
     		free(ptr);
     		ptr = next;
 		}
 		list->current = list->first = list->last = NULL;
 		list->count = 0;
	}
}

void ImgList_Destroy(tImgList lst)
{
	if (lst) {
		ImgList_Clear(lst);
		free(lst);
	}
}

int ImgList_Count(tImgList lst)
{
	ImageList *list	= (ImageList*) lst;
	if (list)
 		return list->count;
 	return 0;
}

int ImgList_Add(tImgList lst, void *ccd)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		list->current = (ImageType*) malloc(sizeof(ImageType));
 		if (list->current) {
	 		memset(list->current, 0, sizeof(ImageType));
 			list->current->image  = ccd;
 			if (list->extra_long>0) {
	 			list->current->lextra = (long*) malloc(list->extra_long*sizeof(long));
	 			if (list->current->lextra) 
	 			    memset(list->current->lextra, 0, list->extra_long*sizeof(long));
 			}
 			if (list->extra_dbl>0) {
	 			list->current->dextra = (double*) malloc(list->extra_dbl*sizeof(double));
	 			if (list->current->dextra) 
	 			    memset(list->current->dextra, 0, list->extra_dbl*sizeof(double));
 			}
 			if (!list->first) 
 				list->last = list->first = list->current;
            else 
             	list->last->next = list->current;
 			list->last = list->current;
 			list->count++;
 			return 1;
		}
	}
	return 0;
}

int ImgList_First(tImgList lst)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		list->current = list->first;
 		return (list->current ? 1 : 0);
	}
	return 0;
}

int ImgList_Next(tImgList lst)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		if (list->current) {
	 		list->current = list->current->next;
     		return (list->current ? 1 : 0);
 		}
 	}
	return 0;
}    

int ImgList_SetImage(tImgList lst, void *ccd)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		if (list->current) {
     		list->current->image  = ccd;
     		return 1;
 		}
   	}
	return 0;
}    

void *ImgList_GetImage(tImgList lst)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		if (list->current) {
 			return list->current->image;
 		}
   	}
	return NULL;
}

int ImgList_SetLong(tImgList lst, int index, long data)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		if (list->current && index>=0 && index<list->extra_long) {
     		list->current->lextra[index] = data;
     		return 1;
 		}
   	}
	return 0;
}    

long ImgList_GetLong(tImgList lst, int index)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		if (list->current && index>=0 && index<list->extra_long) {
 			return list->current->lextra[index];
 		}
   	}
	return 0;
}

int ImgList_SetDouble(tImgList lst, int index, double value)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		if (list->current && index>=0 && index<list->extra_dbl) {
     		list->current->dextra[index] = value;
     		return 1;
 		}
   	}
	return 0;
}    

double ImgList_GetDouble(tImgList lst, int index)
{
	ImageList *list	= (ImageList*) lst;
	if (list) {
 		if (list->current && index>=0 && index<list->extra_dbl) {
 			return list->current->dextra[index];
 		}
   	}
	return 0.0;
}
