/**************************************************************

inifile.c (C-Munipack project)
Copyright (C) 2004 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "comfun.h"
#include "inifile.h"

typedef struct __iniline
{
	char *line;
	char *key;
	char *val;
	char *com;
	struct __iniline *next;
} tiniline;

typedef struct __inifile
{
	struct __iniline *first, *last, *current;
} tinifile;

void iniparseline(const char *buf, char **key, char **val, char **com)
{
	int st;
	const char *bptr;
	char *aux, *aptr;
	size_t keylen, klen, vallen, vlen, comlen, clen;

	aptr = aux = (char*) malloc(strlen(buf)+4);
	memset(aux, 0, strlen(buf)+1);
	for (bptr=buf, st=0; *bptr!=0; bptr++) {
		 switch (st) {
			 case 0: // Skip leading spaces
			 	if ((*bptr>='A' && *bptr<='Z') || (*bptr>='a' && *bptr<='z') || (*bptr=='_')) {
					// Start of keyword
					*aptr=toupper(*bptr); aptr++; st=1;
				} else
				if (*bptr=='#') {
					// Start of comment
					aptr+=2; st=10;
				}
			 	break;
			 case 1: // Reading keyword
			 	if (*bptr=='=' || *bptr<=' ' || *bptr=='#') {
					// End of keyword
					aptr++;
					if (*bptr=='=') st=3;
					else if (*bptr=='#') { aptr++; st=10; }
					else st = 2;
				} else {
					// Keyword continues
					*aptr = toupper(*bptr); aptr++;
				}
				break;
			 case 2: // Spaces before eq. sign
			 	if (*bptr=='=') {
					// Equal sign
					st=3;
				} else
				if (*bptr=='#') {
					// Start of comment
					aptr++; st=10;
				}
			 	break;
			 case 3: // Spaces after eq. sign
			 	if (*bptr=='\'') {
					// Start of 'value'
					st=5;
				} else
			 	if (*bptr=='"') {
					// Start of "value"
					st=7;
				} else
				if (*bptr=='#') {
					// Start of comment
					aptr++; st=10;
				} else
			 	if (*bptr>32) {
					// Start of value (unquoted)
					*aptr=*bptr; aptr++; st=4;
				}
			 	break;
			 case 4: // Reading value (unquoted)
			 	if (*bptr=='#') {
					// Start of comment
					aptr++; st=10;
				} else {
					*aptr=*bptr; aptr++;
				}
			 	break;
			 case 5: // Reading 'value'
			 	if (*bptr=='\'') {
					// Escape character
					st=6;
				} else {
					*aptr=*bptr; aptr++;
				}
			 	break;
			 case 6: // End of 'value' ?
			 	if (*bptr=='\'') {
					// No, single apostophe
					*aptr=*bptr; aptr++; st=4;
				} else {
					// Yes, end of 'value'
					aptr++; st=9;
				}
			 	break;
			 case 7: // Reading "value"
			 	if (*bptr=='"') {
					// Escape character
					st=8;
				} else {
					*aptr=*bptr; aptr++;
				}
			 	break;
			 case 8: // End of "value" ?
			 	if (*bptr=='"') {
					// No, single quote
					*aptr=*bptr; aptr++; st=6;
				} else {
					// Yes, end of "value"
					aptr++; st=9;
				}
			 	break;
			case 9:	// Waiting for sharp sign
				if (*bptr=='#') {
					// Sharp sign
					st=10;
				}
				break;
			case 10:// Spaces after sharp sign
			 	if (*bptr>32) {
					// Start of comment
					*aptr=*bptr; aptr++; st=11;
				}
				break;
			case 11:// Reading comment
				if (*bptr==13 || *bptr==10) {
					// End of comment
					aptr++; st=12;
				} else {
					*aptr=*bptr; aptr++;
				}
				break;
		 }
	}

	// aptr now contains up to three parts (key, value, comment), delimited by null character
	// each part can contain some trailing spaces, which should be removed out. Now, we split
	// them to tree independent strings

	// Skip to key part
	aptr = aux;
	keylen = strlen(aptr);
	if (keylen>0) {
		char *k = aptr + (keylen-1);
	 	while ((k>aptr) && (k[0]<=32)) k--;
	 	k[1] = 0;
		klen = strlen(aptr);
		*key = (char*) malloc(klen+1);
		memcpy(*key, aptr, klen);
		(*key)[klen] = 0;
	} else {
		*key = NULL;
	}
	// Skip to value part
	aptr += keylen+1;
	vallen = strlen(aptr);
	if (vallen>0) {
		char *v = aptr + (vallen-1);
	 	while ((v>aptr) && (v[0]<=32)) v--;
	 	v[1] = 0;
		vlen = strlen(aptr);
		*val = (char*) malloc(vlen+1);
		memcpy(*val, aptr, vlen);
		(*val)[vlen] = 0;
	} else {
		*val = NULL;
	}
	// Skip to comment part
	aptr += vallen+1;
	comlen = strlen(aptr);
	if (vallen>0) {
		char *c = aptr + (comlen-1);
	 	while ((c>aptr) && (c[0]<=32)) c--;
	 	c[1] = 0;
		clen = strlen(aptr);
		*com = (char*) malloc(clen+1);
		memcpy(*com, aptr, clen);
		(*com)[clen] = 0;
	} else {
		*com = NULL;
	}
 	// Normal return
 	//printf("key = '%s'; val = '%s'; com = '%s'\n", *key, *val, *com);
	free(aux);
}

inifile *iniopen(const char *filepath1, const char *filename2)
{
  FILE *f;
  char buf[MAXLINE];
  tinifile *ini;

  /* open file, try filepath1 first */
  if (filepath1 && strlen(filepath1)>0) {
  	f = fopen(filepath1, "rb");
  } else {
#ifdef _WIN32
  	f = fopen(filename2, "rb");
#else
  	char usrname[FILENAME_MAX], usrdir[FILENAME_MAX], etcname[FILENAME_MAX];
  	sprintf(usrname, "%s/.cmunipack/%s", getenv("HOME"), filename2);
	f = fopen(usrname, "rb");
	if (f==NULL) {
		sprintf(usrdir, "%s/.cmunipack", getenv("HOME"));
		if (chkdir(usrdir)) {
			sprintf(etcname, "%s/cmunipack/%s", DATADIR, filename2);
			if (copyfile(etcname, usrname)==0)
				f = fopen(usrname, "rb");
		}
	}
#endif
  }

  /* Check if parameter file was found and open */
  if (f==NULL)
	return NULL;

  /* Allocate inifile record */
  ini = (tinifile*) malloc(sizeof(tinifile));
  ini->first = ini->last = ini->current = NULL;

  /* read all lines into memory */
  while (fgets(buf,MAXLINE,f)!=NULL) {
	if (strlen(buf)>0) {
		ini->current = (tiniline *)malloc(sizeof(tiniline));
		ini->current->line = (char*) malloc(strlen(buf)+1);
		strcpy(ini->current->line, buf);
		iniparseline(ini->current->line, &ini->current->key, &ini->current->val, &ini->current->com);
		if (ini->last) {
			ini->last->next = ini->current;
			ini->last = ini->current;
		} else {
			ini->last = ini->first = ini->current;
		}
		ini->last->next = NULL;
	}
  }
  fclose(f);
  ini->current = ini->first;
  /* Normal return */
  return (inifile *)ini;
}

void iniclos(inifile *ini) {
  tiniline *cur, *aux;
  tinifile *f = (tinifile *)ini;

  if (f) {
	  cur = f->first;
	  while (cur) {
		  if (cur->line) free(cur->line);
		  if (cur->key)  free(cur->key);
		  if (cur->val)  free(cur->val);
		  if (cur->com)  free(cur->com);
		  aux = cur->next;
		  free(cur);
		  cur = aux;
	  }
      free(f);
  }
}

void inirset(inifile *inif)
{
  tinifile *ini = (tinifile *)inif;

  if (ini)
      ini->current = ini->first;
}

int ininkey(inifile *inif)
{
	int count = 0;
  	tinifile *ini = (tinifile *)inif;

  	if (ini) {
    	ini->current = ini->first; 	
 		while (ini->current) {
 			if (ini->current->key)
	 			count++;
			ini->current = ini->current->next;
		}
	}
	return count;
}

int iniread(inifile *inif, char **key, char **val, char **com)
{
  tinifile *ini = (tinifile *)inif;

 if (ini==NULL)
 	return 0;

 while (ini->current) {
 	if (ini->current->key) {
 		if (key) *key = ini->current->key;
		if (val) *val = ini->current->val;
		if (com) *com = ini->current->com;
		ini->current = ini->current->next;
		return 1;
	}
	ini->current = ini->current->next;
 }
 if (key) *key = NULL;
 if (val) *val = NULL;
 if (com) *com = NULL;
 return 0;
}

char *inigkys(inifile *inif, const char *key, const char *def)
{
 char *x, *y;
 tinifile *ini = (tinifile *)inif;

 inirset(ini);
 while (iniread(ini,&x,&y,NULL)) {
     if (strcmp(x, key)==0) {
	 	if (y) {
			char *val = (char*) malloc(strlen(y)+1);
			strcpy(val, y);
			return val;
		} else {
		    char *val = (char*) malloc(1);
			strcpy(val, "");
		    return val;
		}
	}
 }
	if (def) {
		char *val = (char*) malloc(strlen(def)+1);
		strcpy(val, def);
		return val;
	} else {
		char *val = (char*) malloc(1);
		strcpy(val, "");
		return val;
	}
}

double inigkyd(inifile *inif, const char *key, double def)
{
 char *x, *y, *endpos;
 tinifile *ini = (tinifile *)inif;

 inirset(ini);
 while (iniread(ini,&x,&y,NULL)) {
     if (strcmp(x, key)==0) {
	 	if (y) {
			double val = strtod(y, &endpos);
    		return (endpos>y ? val : def);
		} else {
			return def;
		}
	 }
 }
 return def;
}

long inigkyl(inifile *inif, const char *key, long def)
{
 char *x, *y, *endpos;
 tinifile *ini = (tinifile *)inif;

 inirset(ini);
 while (iniread(ini,&x,&y,NULL)) {
     if (strcmp(x, key)==0) {
	 	if (y) {
			long val = strtol(y, &endpos, 10);
    		return (endpos>y ? val : def);
		} else {
			return def;
		}
	 }
 }
 return def;
}
