/**************************************************************

matfun.h (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MATFUN_H
#define MATFUN_H

#define M(x,y) m[(x)+(y)*3]
        // Transformation matrix mapping
#define A(x,y) a[(x)+(y)*5]
        // Mapping for matrix a

extern int    *mmatch_idr1, *mmatch_idr2;
extern int    *mmatch_id1, *mmatch_id2;
extern double *mmatch_r1, *mmatch_r2;
extern double *mmatch_k2, *mmatch_dev;
extern double mmatch_z1, mmatch_z2;
extern double mmatch_x1[MAXSTARS], mmatch_y1[MAXSTARS];       	// Coordinates of stars in reference file
extern double mmatch_x2[MAXSTARS], mmatch_y2[MAXSTARS];       	// Coordinates of stars in input file
extern char   *mmatch_l1[MAXSTARS], *mmatch_l2[MAXSTARS];     	// Text lines of input file
extern int    mmatch_id[MAXSTARS];								// Cross-references (result of Solve procedure)
extern double mmatch_offx, mmatch_offy;							// Frame offsets in pixels (result of Solve procedure)
extern int    mmatch_n1, mmatch_n2, mmatch_c1, mmatch_c2;                    
extern char   *mmatch_line1, *mmatch_line2;                   	// Header lines of input file
extern int    mmatch_nstar;                        				// Max. identification stars 
extern int    mmatch_maxstar;                      				// maximum number of stars used in Solve procedure 
extern double mmatch_clip;                         				// Sigma clipping factor 
extern int    mmatch_spfields;									// Matching mode
extern double mmatch_maxoffset;									// Max. position offset (for Simple method)

void   StInit(void);
       // Initialize stack
void   StAdd(int nstar, double sumsq, double tolerance, int mstar, double *matrix);
       // Add a new member to a stack
void   StSelect(int *n, double *sumsq, double *tolerance, double *matrix);
       // Find the best member 
void   StDump(void);
       // Print content of the stack to debug output
void   StClear(void);
       // Free allocated memory a initialize stack

void   Center(int n, double *x, double *y, double t[2]);
       // Finds a center of a set of points
double Uhel(double a[2], double b[2]);
void   Serad(double a[3]);
void   UV(int i, int j, int k, double *x, double *y, double *u, double *v);
       // Computes u,v lengths of triangle i, j, k */
void   Trafo(double *m, double x1, double y1, double *x2, double *y2);
       // Transform (x1,y1) to (x2,y2) by means of m */ 

#endif
